@@ -10,6 +10,9 @@
 
 #include "mpack/mpack.h"
 
+// We use the Rust implementation for all data transport.
+#define RUST_SOCKETS
+
 //#include <assert.h>
 // By default, python c extensions compile with asserts disable and it is not
 // super easy to disable this behavior (AFAIK). Use this custom assert for now.
@@ -56,7 +59,9 @@ typedef struct {
 // Prototypes
 ////////////////////////////////////////////////////////////////////////////////
 static Writer* Fprofile_Writer(void);
+#if !defined(RUST_SOCKETS)
 static void    Mpack_Flush(Writer* writer, const char* buffer, size_t bytes);
+#endif
 
 static Timestamp Fprofile_Time(void);
 
@@ -117,6 +122,11 @@ static struct {
     // The Unix socket address we're using to communicate with the profile
     // server.
     struct sockaddr_un socket;
+
+#if defined(RUST_SOCKETS)
+    // The Rust extension.
+    PyObject* rust;
+#endif
 } moduleConfiguration;
 
 // The set of methods exposed by this module to Python.
@@ -146,7 +156,7 @@ static PyMethodDef methods[] = {
         .ml_flags = METH_VARARGS,
         .ml_doc   = "INTERNAL USE ONLY"
     },
-    [METHOD_TERMINATE ] = {
+    [METHOD_TERMINATE] = {
         .ml_name  = "terminate",
         .ml_meth  = (PyCFunction) Fprofile_Terminate,
         .ml_flags = METH_NOARGS,
@@ -181,6 +191,7 @@ static Writer* Fprofile_Writer(void) {
     return &state->writer;
 }
 
+#if !defined(RUST_SOCKETS)
 static void Mpack_Flush(Writer* writer, const char* buffer, size_t bytes) {
     // The context passed to Mpack is the socket to send on.
     intptr_t socket = (intptr_t) writer->context;
@@ -190,6 +201,7 @@ static void Mpack_Flush(Writer* writer, const char* buffer, size_t bytes) {
         exit(-1);
     }
 }
+#endif
 
 ////////////////////////////////////////////////////////////////////////////////
 // Misc
@@ -732,6 +744,9 @@ static void Fprofile_ThreadTeardown(void* key) {
     }
 
     mpack_writer_flush_message(&state->writer);
+
+    // TODO: We should tear down the socket to avoid leaking O(# threads)
+    // resources.
 }
 
 static PyObject* Fprofile_MpexitHook(HookedFunction* hook, PyObject* args, PyObject* kwargs) {
@@ -836,10 +851,12 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
         Timestamp   time                      = Fprofile_Time();
         char        traceInitialization[1024] = { 0 };
         Writer      initWriter                = { 0 };
-        intptr_t    sock                      = -1;
         const char* version                   = Py_GetVersion();
         const char* platform                  = Py_GetPlatform();
         char*       outputDirectory           = NULL;
+#if !defined(RUST_SOCKETS)
+        intptr_t    sock                      = -1;
+#endif
 
         // Parse the args to figure out what our output directory is.
         if (!PyArg_ParseTuple(args, "s", &outputDirectory)) {
@@ -848,7 +865,22 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
             exit(-1);
         }
 
-        // Spawn the server and wait for it to connect.
+#if defined(RUST_SOCKETS)
+        // NOTE: We only import this module once.  Importing while tracing is
+        // live will be recorded, which exposes internals users don't care
+        // about and breaks the invariant that we send a `RegisterThread`
+        // message first when threads are traced.
+        moduleConfiguration.rust = PyImport_ImportModule("_functiontrace_rs._lib");
+
+        // Verify that the rust extension loads if we're using it.
+        if (moduleConfiguration.rust == NULL) {
+            perror("Failed to load internal Rust extension");
+            exit(-1);
+        }
+#endif
+
+        // Spawn the server.  We'll wait a bit for it to start accepting
+        // sockets.
         pid_t server = fork();
         if (server == 0) {
             // Detach ourselves as a daemon
@@ -876,6 +908,17 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
                 sizeof(moduleConfiguration.socket.sun_path),
                 "/tmp/functiontrace-server.sock.%d", server);
 
+        // Register an env var so subprocesses know how to connect to our
+        // profile.
+        if (setenv(breadcrumb, (char*) &moduleConfiguration.socket.sun_path, 0)) {
+            perror("Failed to set recursive breadcrumb");
+            exit(-1);
+        }
+
+        // Setup a local Mpack context for this first message.
+        mpack_writer_init(&initWriter, (char*) &traceInitialization, sizeof(traceInitialization));
+
+#if !defined(RUST_SOCKETS)
         // Generate the TraceInitialization message to the profiler server telling
         // them about the profile we'll be creating.
         if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
@@ -902,17 +945,26 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
             usleep(1000);
         }
 
-        // Register an env var so subprocesses know how to connect to our
-        // profile.
-        if (setenv(breadcrumb, (char*) &moduleConfiguration.socket.sun_path, 0)) {
-            perror("Failed to set recursive breadcrumb");
-            exit(-1);
-        }
-
-        // Setup a local Mpack context for this first message.
-        mpack_writer_init(&initWriter, (char*) &traceInitialization, sizeof(traceInitialization));
         mpack_writer_set_context(&initWriter, (void*) sock);
         mpack_writer_set_flush(&initWriter, &Mpack_Flush);
+#else
+        {
+            PyObject* msg_init = PyObject_GetAttrString(moduleConfiguration.rust, "message_initialization");
+            assert(msg_init != NULL);
+
+            PyObject* arg = PyLong_FromVoidPtr(&moduleConfiguration.socket.sun_path);
+            PyObject* context = PyObject_CallFunctionObjArgs(msg_init, arg, NULL);
+            assert(context != NULL);
+
+            PyObject* msg_flush = PyObject_GetAttrString(moduleConfiguration.rust, "message_flush");
+            assert(msg_flush != NULL);
+            PyObject* mpack_flush = PyObject_CallFunctionObjArgs(msg_flush, NULL);
+            assert(mpack_flush != NULL);
+
+            mpack_writer_set_context(&initWriter, PyLong_AsVoidPtr(context));
+            mpack_writer_set_flush(&initWriter, PyLong_AsVoidPtr(mpack_flush));
+        }
+#endif
 
         mpack_start_map(&initWriter, 5);
         {
@@ -947,10 +999,22 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
         // out first.
         mpack_writer_flush_message(&initWriter);
 
+        // Close the initialization socket to trigger `functiontrace-server` to
+        // read our message, since otherwise it'll wait forever for more data.
+#if !defined(RUST_SOCKETS)
         if (shutdown(sock, SHUT_WR) != 0) {
             perror("Failed to close TraceInitialization socket");
             exit(-1);
         }
+#else
+        {
+            PyObject* msg_shutdown = PyObject_GetAttrString(moduleConfiguration.rust, "message_shutdown");
+            assert(msg_shutdown != NULL);
+
+            PyObject* socket = PyLong_FromVoidPtr(mpack_writer_context(&initWriter));
+            assert(PyObject_CallFunctionObjArgs(msg_shutdown, socket, NULL) != NULL);
+        }
+#endif
     } else {
         // Read the socket addr out of `existingProfile` so our threads know
         // how to talk to the profile server.
@@ -1033,6 +1097,14 @@ static ThreadState* Fprofile_CreateThreadState(void) {
         exit(-1);
     }
 
+    // Setup msgpack and a socket for this thread to report tracing info on.
+    //
+    // We need to do this before any allocations or Python function calls
+    // occur, since consumers can't tell whether mpack is fully initialized or
+    // not.
+    mpack_writer_init(&state->writer, (char*) &state->buffer, sizeof(state->buffer));
+
+#if !defined(RUST_SOCKETS)
     // Allocate a socket to report this thread's information back to the
     // profile server.
     if ((state->socket = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
@@ -1046,11 +1118,33 @@ static ThreadState* Fprofile_CreateThreadState(void) {
         exit(-1);
     }
 
-    // Register a messagepack writer for this thread, including telling it to
-    // flush on this thread's socket once the buffer fills up.
-    mpack_writer_init(&state->writer, (char*) &state->buffer, sizeof(state->buffer));
+    // Configure Mpack to flush events to this thread's socket once the buffer
+    // fills up.
     mpack_writer_set_context(&state->writer, (void*) state->socket);
     mpack_writer_set_flush(&state->writer, &Mpack_Flush);
+#else
+    {
+        PyObject* msg_init = PyObject_GetAttrString(moduleConfiguration.rust, "message_initialization");
+        assert(msg_init != NULL);
+
+        // TODO: Convert this to a shared Rust configuration object so we don't
+        // need to pass it back and forth.
+        PyObject* arg = PyLong_FromVoidPtr(&moduleConfiguration.socket.sun_path);
+        PyObject* context = PyObject_CallFunctionObjArgs(msg_init, arg, NULL);
+        assert(context != NULL);
+
+        PyObject* msg_flush = PyObject_GetAttrString(moduleConfiguration.rust, "message_flush");
+        assert(msg_flush != NULL);
+        PyObject* mpack_flush = PyObject_CallFunctionObjArgs(msg_flush, NULL);
+        assert(mpack_flush != NULL);
+
+        // Re-initialize msgpack, since we've done some allocations and Python
+        // calls that we don't want to actually record.
+        mpack_writer_init(&state->writer, (char*) &state->buffer, sizeof(state->buffer));
+        mpack_writer_set_context(&state->writer, PyLong_AsVoidPtr(context));
+        mpack_writer_set_flush(&state->writer, PyLong_AsVoidPtr(mpack_flush));
+    }
+#endif
 
     // Send a message about our thread's registration.
     Timestamp time = Fprofile_Time();
