@@ -41,7 +41,7 @@ def setup_dependencies():
     # Generate wrappers for the various Python versions we support to ensure
     # they're included in our PATH.
     wrap_pythons = ["python", "python3"] + [
-        "python3.{}".format(minor) for minor in range(6, 12)
+        "python3.{}".format(minor) for minor in range(6, 13)
     ]
     for python in wrap_pythons:
         with open(os.path.join(tempdir, python), "w") as f:
@@ -55,9 +55,7 @@ def main():
         "--trace-memory",
         "--trace_memory",
         action="store_true",
-        help="""Trace
-                        memory allocations/frees when enabled.  This may add
-                        tracing overhead, so is disabled by default.""",
+        help="Trace memory allocations/frees when enabled.  This may add tracing overhead, so is disabled by default.",
     )
     parser.add_argument(
         "--output-dir",
@@ -66,6 +64,12 @@ def main():
         default=os.getcwd(),
         help="The directory to output trace files to",
     )
+    parser.add_argument(
+        "--no-compression",
+        "--no_compression",
+        action="store_true",
+        help="Disable compression of traces, which can be unnecessarily slow if filesize is not a concern.",
+    )
     parser.add_argument(
         "-v", "--version", action="version", version=_functiontrace.__version__
     )
@@ -107,6 +111,8 @@ def main():
     setup_dependencies()
 
     # Setup our tracing environment, including configuring tracing features.
+    if args.no_compression:
+        os.environ["FUNCTIONTRACE_COMPRESSION"] = "false"
     if args.trace_memory:
         _functiontrace.config_tracememory()
     _functiontrace.begin_tracing(args.output_dir)
